require 'entities.bullet'

Ship = class{name = 'Ship', function(self, x,y, texture, endurance, speed, agility, accuracy, gun_delay)
	self.texture = texture and IMAGES[texture] or IMAGES['ship.png']
	local w,h = self.texture:getWidth()/2, self.texture:getHeight()
	self.quads = {
		idle     = love.graphics.newQuad(0,0, w,h, 2*w, h),
		shooting = love.graphics.newQuad(w,0, w,h, 2*w, h),
	}
	self.state = 'idle'
	self.may_shoot = true
	self.gun_delay = gun_delay or .5
	self.ox = w/2
	self.oy = h/2

	self.shape = HC.addRectangle(x-w/2, y-h/2, w,h)
	HC.addToGroup(self, self.shape)
	self.shape.class = Ship
	self.shape.self = self

	self.vel = vector(0,0)
	self.last_vel = vector(0,0)

	self.life = 1
	self.endurance = endurance or 1

	self.speed = speed or 10
	self.agility = agility or 1
	self.accuracy = accuracy or .1
	self.field_of_view = 600

	self.on_shoot = function() end
	self.on_die = function() end

	local p = love.graphics.newParticleSystem(IMAGES['splash.png'], 512)
	p:setEmissionRate(256)
	p:setOffset(4,4)
	p:setParticleLife(.8,1.2)
	p:setSpeed(10,40)
	p:setSpread(math.pi*2)
	p:start()
	self.particles = p

	local function random_goal()
		self.random_goal = self:pos() + vector(0,1):rotate_inplace(math.random() * math.pi * 2) * 500
	end
	Timer.add(.1, random_goal)
	Timer.addPeriodic(15, random_goal)
	self.random_goal = vector(x,y)

	-- register oneself
	ENTITIES[self] = self
end}

function Ship:draw()
	local x,y = self.shape:center()
	love.graphics.setColor(255,255,255)
	love.graphics.draw(self.particles)
	love.graphics.drawq(self.texture, self.quads[self.state],
		x,y, self.shape:rotation(), 1,1, self.ox, self.oy)

	-- life bar
	love.graphics.setColor(200,30,20,128)
	love.graphics.rectangle('fill', x-30, y-self.oy - 10, 60, 5)
	love.graphics.setColor(30,200,30)
	love.graphics.rectangle('fill', x-30, y-self.oy - 10, self.life*60, 5)
end

function Ship:shoot(group)
	if not self.may_shoot then return end
	
	local jitter = (math.random() * 2 - 1) * self.accuracy
	local up = vector(0,-1):rotate_inplace( self.shape:rotation() + jitter )
	self.state = 'shooting'
	Timer.add(.1, function() self.state = 'idle' end)
	Timer.add(self.gun_delay, function() self.may_shoot = true end)
	self.may_shoot = false

	local pos = self:pos() + up * 24
	Bullet(pos, up * 800, 2, group)

	self:on_shoot( up )
end

function Ship:move(dt)
	self.vel = self.vel * self.speed
	self.vel = self.vel + self.last_vel * .92
	self.last_vel = self.vel
	self.shape:move( (self.vel * dt):unpack() )

	self.particles:setPosition(self.shape:center())
	self.particles:update(dt)
end

local function absmin(a,b)
	return math.abs(a) <= math.abs(b) and a or b
end
local function sign(x)
	if x < 0 then return -1 end
	return x > 0 and 1 or 0
end

local function drive_to(dt, self, pos, speed)
	local to = pos - self:pos()
	if to:len() < 10 then self.vel = vector(0,0) end
	local up = vector(0,-1):rotate_inplace( self.shape:rotation() )

	local angle = math.atan2(to.y,to.x) - math.atan2(up.y,up.x)
	self.shape:rotate( self.agility * sign(angle) * dt )

	if math.abs(angle) < math.pi / 4 then
		self.vel = up:rotated( angle * dt ) * (speed or 1)
	else
		self.vel = vector(0,0)
	end

	return angle
end
function Ship:update(dt)
	local toplayer = (self.player:pos() - self:pos())
	if toplayer:len() < self.field_of_view then
		local angle = drive_to(dt, self, self.player:pos())
		if math.abs(angle) < math.pi / 16 then
			self:shoot(self)
		end
	else
		drive_to(dt, self, self.random_goal, .3)
	end

	self:move(dt)
end

function Ship:destroy()
	audio.play(SOUNDS['explode1.ogg'])
	HC.remove(self.shape)
	ENTITIES[self] = nil
end

function Ship:pos()
	return vector(self.shape:center())
end

Ship.collision_handler = {}
Ship.collision_handler[Ship] = function(self, other, dx,dy)
	self.shape:move(dx/2, dy/2)
	other.shape:move(-dx/2, -dy/2)
end

Ship.collision_handler[Bullet] = function(self, bullet)
	self.life = self.life - math.max(bullet.damage * self.endurance, 0)
	if self.life < 0 then
		self:on_die()
		self:destroy()
	end

	bullet:destroy()
end

package.loaded[...] = Ship
