local state = GS.new()
package.loaded[...] = state

-- entity table that forwards every method call to
-- all entries in the table
ENTITIES = setmetatable({}, {__index = function(t, key)
	return function(...)
		for _,e in pairs(t) do e[key](e, ...) end
	end
end})

require 'entities.ship'
require 'entities.player'
require 'entities.platform'
require 'entities.bullet'

-- ----------------------- --
--     STATE FUNCTIONS     --
-- ----------------------- --
local shaker
local BG_TEX_WIDTH
local BG_TEX_HEIGHT
ENEMIES = {}
function state:init()
	BG_TEX_WIDTH = 1000 * SCR_WIDTH
	BG_TEX_HEIGHT = 1000 * SCR_HEIGHT
	self.quad_bg = love.graphics.newQuad(0,0, BG_TEX_WIDTH, BG_TEX_HEIGHT, 512,512)

	shaker = Timer.Oscillator(2.5, function(s) return math.sin(s * 2 * math.pi) * .3 end)

	IMAGES['ship.png']:setFilter('nearest', 'nearest')
	IMAGES['player.png']:setFilter('nearest', 'nearest')
	IMAGES['bullet.png']:setFilter('nearest', 'nearest')
	IMAGES['explosion.png']:setFilter('nearest', 'nearest')
end

function state:enter()
	HC.init(100)
	HC.setCallbacks{persist = function(_, a,b, dx,dy)
		if a.class.collision_handler and a.class.collision_handler[b.class] then
			return a.class.collision_handler[b.class](a.self,b.self, dx,dy)
		elseif b.class.collision_handler and b.class.collision_handler[a.class] then
			return b.class.collision_handler[a.class](b.self,a.self, -dx,-dy)
		end
	end}

	self.cam = camera()
	self.player = Player( SCR_WIDTH/2,SCR_HEIGHT/2 )
	HC.addToGroup(Player, self.player.shape)

	function self.player:on_die()
		GS.switch(states.outro, 'loser')
	end

	ENEMIES = {}
	for i = 1,5 do
		local x,y = (math.random() * 2 - 1) * 5 * SCR_WIDTH, (math.random() * 2 - 1) * 5 * SCR_HEIGHT
		Platform(x,y, self.player)
	end

	audio.play(SOUNDS['water.ogg'], .5, true)
	audio.play(SOUNDS['boat.ogg'], .2, true)
end

function state:leave()
	audio.stopAll()
end

function state:draw()
	self.cam:predraw()
	love.graphics.setColor(255,255,255)
	love.graphics.drawq(IMAGES['water03.jpg'], self.quad_bg, -BG_TEX_WIDTH/2, -BG_TEX_HEIGHT/2)
	ENTITIES.draw()
	self.cam:postdraw()

	-- HUD
	love.graphics.setColor(255,255,255)
	love.graphics.setFont(FONTS[30])
	love.graphics.draw(IMAGES['cash.png'], 10,10)
	love.graphics.print(self.player.cash, 50, 5)

	love.graphics.setColor(255,30,30)
	local p = self.player:pos()
	local r = self.cam:toCameraCoords(p)
	local has_enemy = false
	for _,e in pairs(ENEMIES) do
		has_enemy = true
		local q = vector(e.shape:center())
		q = (q - p)
		local dist = q:len()
		q = q / dist
		if dist > 600 then
			love.graphics.line(r.x+q.x*200, r.y+q.y*200, (r + q*250):unpack())
		end
	end

	if not has_enemy then
		GS.switch(states.outro, 'winner')
	end
end

function state:update(dt)
	dt = math.min(dt, 1 / 25) -- min 25 fps

	ENTITIES.update(dt)

	local ppos = self.player:pos()
	self.cam:translate( (ppos - self.cam.pos) * .5 )

	local d = shaker(dt)
	local speed = self.player.vel:len()
	self.cam:translate( 5 * vector(d,d) )
	self.cam.zoom = .8 + d/40 - speed / 1500

	HC.update(dt)

	-- you fell off the world
	if math.abs(ppos.x) > BG_TEX_WIDTH/2 or math.abs(ppos.y) > BG_TEX_HEIGHT/2 then
		love.graphics.setBackgroundColor(0,0,0)
		local sd = state.draw
		local su = state.update
		local sk = state.keypressed
		function state:draw()
			love.graphics.setColor(255,255,255)
			love.graphics.setFont(FONTS[50])
			local str = 'You fell off the world.\nConratulations.\nYou broke the game.\nI hope you are happy now.'
			love.graphics.printf(str, 0,(SCR_HEIGHT - FONTS[50]:getHeight(str))/2 - 100, SCR_WIDTH, 'center')
		end

		function state:update()
		end

		local count = 0
		function state:keypressed()
			count = count + 1
			if count > math.random(5,10) then
				state.draw = sd
				state.update = su
				state.keypressed = sk
				GS.switch(states.menu)
			end
		end
	end
end

function state:keyreleased(key)
	if key == 'p' then
		local update = love.update
		local draw = love.draw
		local keyreleased = love.keyreleased

		function love.update() end
		function love.draw()
			draw()
			love.graphics.setColor(0,0,0,100)
			love.graphics.rectangle('fill', 0,0, SCR_WIDTH, SCR_HEIGHT)
			love.graphics.setColor(255,255,255)
			love.graphics.setFont(FONTS[50])
			love.graphics.printf('PAUSE', 0,SCR_HEIGHT/2-FONTS[50]:getHeight('PAUSE'), SCR_WIDTH, 'center')
		end
		function love.keyreleased(key)
			if key == 'p' then
				love.update = update
				love.draw = draw
				love.keyreleased = keyreleased
			end
		end
	end
end

return state
