local Shapes = require 'hardoncollider.shapes'

local state = GS.new()
package.loaded[...] = state

-- simple button
Button = class{name = 'Button', function(self, label, x,y, w,h, font)
	local w = w or 400
	local h = h or 40
	self.shape = Shapes.PolygonShape(x,y, x+w,y, x+w,y+h, x,y+h)

	-- button label
	self.label = assert(label, "Button has no label")
	self.font = font or FONTS[20]

	-- button alignment
	self.labelx = x + (w - self.font:getWidth(label)) / 2
	self.labely = y + (h - self.font:getHeight(label)) / 2

	self.highlight = false
end}

function Button:draw()
	local border, bg, text = {0,0,0}, {255,255,255}, {0,0,0}
	if self.active then
		border = {200,200,200}
		text = border
		bg = {40,40,40}
	end

	-- button border
	love.graphics.setColor(unpack(border))
	self.shape:draw('fill')

	-- button background
	love.graphics.setColor(unpack(bg))
	local x1,y1, x2,y2 = self.shape._polygon:getBBox()
	love.graphics.rectangle('fill', x1+4,y1+4, x2-x1-8,y2-y1-8)

	-- label
	love.graphics.setColor(unpack(text))
	love.graphics.setFont(self.font)
	love.graphics.print(self.label, self.labelx, self.labely)
end

function Button:contains(x,y)
	return self.shape:contains(x,y)
end

-- ----------------------- --
--     STATE FUNCTIONS     --
-- ----------------------- --
function state:init()
	local buttons_y = 450
	self.buttons = {
		start   = Button('Start', SCR_WIDTH/2 - 200,   buttons_y),
		options = Button('Options', SCR_WIDTH/2 - 200, buttons_y + 60),
		quit    = Button('Quit', SCR_WIDTH/2 - 200,    buttons_y + 120),
	}

	-- button actions
	function self.buttons.start.command()
		GS.switch(states.intro, true)
	end

	function self.buttons.options.command()
		-- blank screen for some time
		local draw = self.draw
		local update = self.update
		local keyreleased = self.keyreleased
		local mousereleased = self.mousereleased
		local bg = {love.graphics.getBackgroundColor()}
		love.graphics.setBackgroundColor(0,0,0)
		Timer.add(math.random(.6,1.5), function()
			love.graphics.setBackgroundColor(bg)
			self.draw          = draw
			self.update        = update
			self.keyreleased   = keyreleased
			self.mousereleased = mousereleased
		end)
		self.draw          = function() end
		self.update        = function() end
		self.keyreleased   = function() end
		self.mousereleased = function() end
	end

	function self.buttons.quit.command()
		love.event.push('q')
	end

	-- wire buttons
	self.buttons.start.next = self.buttons.options
	self.buttons.start.prev = self.buttons.quit
	self.buttons.options.next = self.buttons.quit
	self.buttons.options.prev = self.buttons.start
	self.buttons.quit.next = self.buttons.start
	self.buttons.quit.prev = self.buttons.options

	self.quad_bg = love.graphics.newQuad(0,0, SCR_WIDTH * 2, SCR_HEIGHT * 2, 512,512)
	self.bg_pos = Timer.Oscillator(10, function(f)
		local t = f * 2 * math.pi
		return (math.sin(t) - 1)/2 * SCR_WIDTH, (math.cos(t) - 1)/2 * SCR_HEIGHT
	end)
end

function state:enter(pre)
	self.selected = self.buttons.start

	if pre ~= states.outro then
		audio.stream('pharmacokinetik.mp3', .6)
	end
end

local bx,by = 0,0
local function centerText(text, y, size)
	local fw = FONTS[size]:getWidth(text)
	love.graphics.print(text, (SCR_WIDTH - fw)/2, y)
end
function state:draw()
	love.graphics.setColor(255,255,255)
	love.graphics.drawq(IMAGES['water01.jpg'], self.quad_bg, bx,by)

	love.graphics.setFont(FONTS[70])
	centerText('130 A.F.', 100, 70)

	love.graphics.setFont(FONTS[15])
	centerText('a game by vrld - the worst one yet', 170, 15)

	love.graphics.print('made with', SCR_WIDTH - 123, SCR_HEIGHT - 32)
	love.graphics.draw(IMAGES['love_ball.png'], SCR_WIDTH - 40, SCR_HEIGHT - 40, 0, .5,.5)

	-- copyright message
	local cpy_off = 240
	love.graphics.setFont(FONTS[20])
	centerText('awesome music by Binrpilot:', cpy_off, 20)

	love.graphics.setFont(FONTS[15])
	centerText('Remember C64?//Pharmacokinetik', cpy_off + 27, 15)
	love.graphics.setFont(FONTS[12])
	centerText('(go to www.binaerpilot.no to show your appreciation)', cpy_off + 50, 12)

	for _,button in pairs(self.buttons) do
		button.active = button == self.selected
		button:draw()
	end
end

function state:keyreleased(key)
	if key == 'return' or key == ' ' then
		-- probably invoke button command
		if math.random() < .99 then
			self.selected.command()
		end
	end
end

function state:update(dt)
	bx,by = self.bg_pos(dt)
	-- make it hard to navigate the menu with the keyboard
	if love.keyboard.isDown('up') then
		self.selected = self.selected.prev
		audio.play(SOUNDS['mv.ogg'])
	elseif love.keyboard.isDown('down') then
		self.selected = self.selected.next
		audio.play(SOUNDS['mv.ogg'])
	end

	for _,button in pairs(self.buttons) do
		if button:contains(love.mouse.getPosition()) then
			if self.selected ~= button then
				audio.play(SOUNDS['mv.ogg'])
			end
			self.selected = button
		end
	end

end

function state:mousereleased(x,y, btn)
	if btn == 'l' then
		if self.selected:contains(x,y) and math.random() < .99 then
			self.selected.command()
		end
	end
end

return state
